/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#ifndef BASICSOLVER_HPP_
#define BASICSOLVER_HPP_

#include	<vector>
#include	"ConstraintDB.hpp"
#include	"Types.hpp"
#include	"Settings.hpp"
#include	"Statistics.hpp"
#include	"DecayHeap.hpp"
#include	"BitVector.hpp"

using		std::vector;

//Specify what to do with variable values when they are undone
enum BackjumpType{BACKJUMP_SET_PREV,BACKJUMP_SET_PREV_FLP,BACKJUMP_PREV_UNTOUCHED};


/**
 * Class to implement the main solver
 */
class BasicSolver: public ConstraintDB{

	//return value of boolean constraint propagation
	static const int32_t BCP_SUCCESS = -1; //SOMETHING < 0

	/** A heuristic measurement of the activity of a variable  */
	DecayHeap< Variable, var_act_t >   	var_act_hp;

	/** Indicates if state of the solver is ok, or if formula is already unsatisfied */
	bool								state_ok;

	/** Assignment stack; stores all assignments made in the order they were made*/
	vector<Literal>				 		assign_stack;

	/** Head of queue - as index into the assign_stack */
	uint32_t							queue_begin;

	/** Indices for different decision levels in 'assign_stack'
	 * decision_level_end[l] states that all assignments in assign_stack[0,l) are made above or at decision level l*/
	vector<uint32_t>					decision_level_end;

	/** Number of top-level assignments at last execution of 'simplify()' */
	uint32_t							simplify_assign_idx;

	/** Remaining number of propagations that must be made before next execution of 'simplify()' is allowed to run*/
	int64_t						 		simplify_props_remaining;

	/** Data structure to set a flag for a variable and allow efficient clearing of all flags */
	BitVector<Variable>			 		var_flags;

	/**Timestamp for the caching of indices into clauses, when a new watcher is searched*/
	uint32_t							clause_stamp;

	//current settings
	double  							curr_max_conflicts;
	double  							curr_max_learnts;


	//conflict related information
	vector<Literal>						conflict_lemma;
	uint64_t							conflict_involved_levels;

	//variables whose flag is set
	vector<Variable>					var_flags_to_clear;

	//space to save some literals (for instance in firstDecisionCheckBothPol(), reduceSelfImpl())
	vector<Literal>				 		tmp_literal_store;



 //Methods to simplify the clause database
	void 			clearShortWatchers			(Variable var);
	void 			clearShortWatchers			(Literal lit);
	void			simplifyAllShortWatchers 	();
	void			simplifyShortWatchersOf 	(vector<WatchedElement>& wes);
	TrBool		  	simplify					( );
	TrBool 			removeSatAndShrinkClauses	(vector<Clause*>& clses,uint32_t max_cls_sz);
	uint32_t 		clsCheckSatAndShrink		(Clause& cl);
	uint32_t 		clsCheckSatAndShrinkConsAct	(Clause& cl);
	TrBool 			bigSimplify					( );

//unit propagation
	void			firstDecisionCheckBothPol(Literal next);
	int32_t			bcp		   			(const bool init_conflict );
	int32_t			bcpPropagateLit		(const bool init_conflict, const Literal& false_lit);
	inline void 	bcpFinishOnConflict	(we_vec&  watched, we_iter iter, we_iter copy,we_iter end);
	inline int32_t	bcpInitConflict		(const int32_t& dec_level, const Literal& w1, const Literal& w2);
	inline bool		bcpFindNewWatcher	(Clause& cl, const Literal& false_lit, const uint32_t& other_lit_idx);

//learning
	inline void		backjumpKeepLevel   (uint32_t level,const BackjumpType type);
	void			backjumpUndoAssigns (uint32_t assign_idx, const BackjumpType type); //undoes all assignments above and including the specified assign_idx
	bool			firstUIP			(int32_t nbr_pivot_lits);
	inline void		firstUIPCheckLiteral(const Literal& lit,const int32_t& dec_level,int32_t& nbr_pivot_elements);

	uint32_t		backtrackLevel		();//decides to which level we have to backtrack
	void			reduceSelfImpl		();
	inline void		unsetConfLemmaFlags	();
	bool			litIsNecessary	  	(Literal lit);
	inline bool		reduceSelfImpCheckLiteral(const Literal& lit);

//heuristic
	inline void		varBump				(const Variable& var);
	inline void		claDecayActivity	();

//solving
	inline void		reinsertVarOrder	(Variable v);
	inline void		newDecisionLevel	();
	template<class Reason>
	inline void		uncheckedEnqueue	(const Literal& lit, Reason from = (Clause*)NULL);
	inline void		uncheckedEnqueue	(const Literal& lit, const Literal& r1,const Literal& r2 );
	inline uint32_t numberAssignments()	const;
	inline uint32_t decisionLevel()		const;
	Literal			nextDecision();
	TrBool			search();
	bool			addValidClause(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend);

	//sort literals by the activity values of their variables
	struct LitActDescendingFunctor{
		BasicSolver* outer ;
		LitActDescendingFunctor(BasicSolver* s):outer(s){};
	  	bool operator() (const Literal& a, const Literal& b){return outer->var_act_hp.getValue(a.toVar()) > outer->var_act_hp.getValue(b.toVar());	}
	};

public:
					BasicSolver(Settings& sets, Statistics& stats);
					~BasicSolver();
	inline void		setInvalid();
	inline void		reserveForVariables(uint32_t sz);
			bool	addClause(vector<Literal>& lits);
			TrBool  isSAT(double timeout = -1);

};



//Inline Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
void	BasicSolver::setInvalid()	{state_ok = false;}

void	BasicSolver::reinsertVarOrder(Variable v){
	if (!var_act_hp.contains(v)) var_act_hp.restoreElement(v);
}


void	BasicSolver::varBump(const Variable& var){
	if(!var_act_hp.bump(var)) ++statistics.nbr_expensive_var_bmp;
}



template<class Reason>
void	BasicSolver::uncheckedEnqueue(const Literal& lit, Reason from){
	assert(litIsUnknown(lit));
	varUncheckedAssign(lit,from,decisionLevel());
	assign_stack.push_back(lit);
}

void	BasicSolver::uncheckedEnqueue(const Literal& lit, const Literal& r1,const Literal& r2 ){
	assert(litIsUnknown(lit));
	assert(litIsFalse(r1) && litIsFalse(r2));
	varUncheckedAssign(lit,r1,r2,decisionLevel());
	assign_stack.push_back(lit);
}


void		BasicSolver::claDecayActivity()				{cls_inc *= settings.cls_decay; }

uint32_t	BasicSolver::decisionLevel ()		const	{return decision_level_end.size(); }

uint32_t	BasicSolver::numberAssignments()	const	{return assign_stack.size();}

void		BasicSolver::newDecisionLevel()				{decision_level_end.push_back(assign_stack.size()); }



void	 	BasicSolver::reserveForVariables(uint32_t sz){
	if(sz > numberVariables()){
		ConstraintDB::reserveForVariables(sz);
		var_act_hp.reserve(sz);
	}
}



void		BasicSolver::backjumpKeepLevel(uint32_t level,BackjumpType type){

	if(decisionLevel() <= level) return;
	backjumpUndoAssigns(decision_level_end[level],type);
	decision_level_end.resize(level);
}


void BasicSolver::bcpFinishOnConflict(we_vec&  watched, we_iter iter, we_iter copy,we_iter end){
	queue_begin 	= assign_stack.size(); // to finish outer while loop in bcp()
	while(iter != end) *copy++ = *iter++;
	if(copy != end) watched.erase(copy,end);
}

//helper function
//initialise the data structure to analyse the conflict in FirstUIP
int32_t	BasicSolver::bcpInitConflict(const int32_t& dec_level, const Literal& w1, const Literal& w2){
	assert(var_flags.allCleared());

	conflict_lemma					.resize(1,INVALID_LIT);
	conflict_involved_levels 		= 0;
	int32_t nbr_pivot_lits 			= 0;
	firstUIPCheckLiteral(w1,dec_level,nbr_pivot_lits);
	firstUIPCheckLiteral(w2,dec_level,nbr_pivot_lits);
	return nbr_pivot_lits;
}


//helper function
//try to find a new watcher if one watching literal is false
bool	BasicSolver::bcpFindNewWatcher(Clause& cl, const Literal& false_lit, const uint32_t& other_lit_idx){
	uint32_t this_lit_idx = other_lit_idx^1;

	for(uint32_t k=cl.getNextK(clause_stamp);k<cl.size();++k){
		if(litIsNotFalse(cl[k])){
			cl[this_lit_idx] = cl[k];
			cl[k]			= false_lit;
			cl.setNextK(k+1,clause_stamp);

			//we is "deleted" (not copied) from the watchers list
			litAddWatchedClause(cl[this_lit_idx],cl[other_lit_idx],&cl);
			return true;
		}
	}
	return false;
}


//helper function
//go back the implication graph
void BasicSolver::firstUIPCheckLiteral(const Literal& lit,const int32_t& dec_level,int& nbr_pivot_elements){
   const Variable	  var(lit.toVar());

   if(var_flags.isSet(var)) return;
   assert(litIsFalse(lit));

   int level = varAssignLevel(var);
   if(level > 0){
	   var_flags			   .setFlag(var);
	   varBump(var);

	   if(level <  dec_level){
		   conflict_lemma			.push_back(lit);
		   conflict_involved_levels |= hash64(level);
	   }else {
		   ++nbr_pivot_elements;
	   }
   }
}

//clear all variable flags
void BasicSolver::unsetConfLemmaFlags(){
	uint32_t cp = conflict_lemma.size();
	while(cp--) {var_flags.carelessClear(conflict_lemma[cp].toVar());}
	for(cp = var_flags_to_clear.size();cp--;) {var_flags.carelessClear(var_flags_to_clear[cp]);}
	var_flags_to_clear.clear();
	assert(var_flags.allCleared());
}

//helper function for self subsumption
//resturns true if the given literal is necessary for the conflict clause and search
//cannot continue recursively with the assign reason for this lit
bool BasicSolver::reduceSelfImpCheckLiteral(const Literal& lit){

	Variable var = lit.toVar();

	if(var_flags.isSet(var)) return false;

	const AssignInfo& a_info = varAssignInfo(var);
	if(a_info.getLevel() > 0){
		if(a_info.noReason() || ((hash64(a_info.getLevel()) & conflict_involved_levels) == 0)){
			return true;
		}

		var_flags.setFlag(var);
		tmp_literal_store.push_back(lit);
	}
	return false;
}



#define for_all_but_one(idx,sz,todo) {todo}; for(idx = 2;idx < sz;++idx){ todo}


const char*  basic_solver_cpp_version();
const char*  filtered_solver_cpp_version();
const char*  sapperlot_cpp_version();
#define basic_solver_hpp_version "BasicSolver.hpp:	$Revision: 1.15 $"


























#endif /*BASICSOLVER_HPP_*/
