/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#include "ConstraintDB.hpp"
#include "Settings.hpp"

//binary clauses are stored directly in the watchers list
void ConstraintDB::clsCreateBinary(const Literal lit0, const Literal lit1){

    assert(lit0.idx() < watches.size() && lit1.idx() < watches.size());

    watches[lit0.idx()]	.push_back(WatchedElement(lit1));
   	watches[lit1.idx()]	.push_back(WatchedElement(lit0));
   	nbr_short_literals	+= 2;
   	++nbr_bin_clses;
}

//ternary clauses are stored in the watchers lists of all three literals
void ConstraintDB::clsCreateTernary(const Literal lit0, const Literal lit1,const Literal lit2){

    assert(lit0.idx() < watches.size() && lit1.idx() < watches.size() && lit2.idx() < watches.size());

	watches[lit0.idx()]	.push_back(WatchedElement(lit1,lit2));
	watches[lit1.idx()]	.push_back(WatchedElement(lit0,lit2));
	watches[lit2.idx()]	.push_back(WatchedElement(lit0,lit1));
	nbr_short_literals	+= 3;
	++nbr_tern_clses;
}

//create a clause but init the memory from outside
Clause* ConstraintDB::clsCreate(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend, bool learnt){

    uint32_t cl_sz							= lend-lbeg;
    assert(cl_sz > 3);

    void* mem								= malloc(sizeof(Clause) + sizeof(Literal)*cl_sz);
    if(mem == NULL){
    	fprintf(stderr,"\nc !!!OUT OF MEMORY!!!\n");
    	exit(EXIT_FAILURE);
    }

    Clause* cls 							= new (mem) Clause(cl_sz, learnt);

    Literal lit0 							= *lbeg;
    Literal lit1 							= *(lbeg+1);

    for(uint32_t i = 0;lbeg != lend;++lbeg) {(*cls)[i++] = *lbeg;}

    if(learnt)  {
        learnt_cls                          .push_back(cls);
        nbr_long_learnt_literals			+= cl_sz;
    }else{
        fix_cls                             .push_back(cls);
        nbr_long_fix_literals				+= cl_sz;
    }

    clsLinkWatched(cls);

    return cls;
}


//free a clause's memory and unlink the clause from the watcher lists
void    ConstraintDB::clsUncreate(Clause* cl,bool unlink){
    assert(cl->size() > 1);

    if(cl->isLearnt())	nbr_long_learnt_literals 	-= cl->size();
    else				nbr_long_fix_literals 		-= cl->size();

    if(unlink) {clsUnlinkWatched(cl);}

    free(cl);
}



//ensure that data structures have enough space for all variables and lits
void ConstraintDB::reserveForVariables(uint32_t need_sz){

    if(need_sz > assignments.size()){
        //Variable infos
        assignments     .resize(need_sz,UNKNOWN_TrBool.toInt());
        assign_info     .resize(need_sz,AssignInfo());

        //Literal infos
        need_sz 		= need_sz << 1;
        watches         .resize(need_sz);
    }
}

//shrink the vector of learnt clauses to the particular size
void ConstraintDB::learntsGC(uint32_t new_sz){

	if(new_sz >= learnt_cls.size()){return;}

    statistics.nbr_learnt_reductions += 1;

    std::nth_element(learnt_cls.begin(), learnt_cls.begin() + new_sz - 1,
                       learnt_cls.end(), Clause::ActivityDescendingFunctor());

    for(uint32_t del_idx = new_sz; del_idx < learnt_cls.size();++del_idx){
    	 if(clsLocked(*(learnt_cls[del_idx])))   learnt_cls[new_sz++] = learnt_cls[del_idx];
    	 else                                    clsUncreate(learnt_cls[del_idx]);
    }

    learnt_cls.resize(new_sz);
}


//check if clause is already satisfied
//also remove literals that are false anyway
bool ConstraintDB::clsCanBeOmitted(Clause& cl){


	for(uint32_t i = 0; i != cl.size(); ) {
		if(litIsTrue(cl[i])){ 							return true;}
		if(litIsFalse(cl[i]) && i > 1){ // if i <= 1 -> clause is unit and other watcher should be true
			clsRemoveLiteral(cl,i);
			continue;
		}
		++i;
	}

	const uint32_t sz = cl.size(); assert(sz > 1);

	if(sz > 3) 											return false;

	if(sz == 3){clsCreateTernary(cl[0],cl[1],cl[2]); 	return true;}

	clsCreateBinary(cl[0],cl[1]);						return true;
}

//remove all satisfied clauses from the given vector
void ConstraintDB::removeSat(vector<Clause*>& cls){
    vector<Clause*>::iterator to, iter;
    for (iter = to = cls.begin(); iter != cls.end(); ++iter){
        if (clsCanBeOmitted(**iter)){
        	statistics.nbr_simplify_worked += 1;
            clsUncreate(*iter);
        }else{
        	assert(litIsUnknown((**iter)[0])&& litIsUnknown((**iter)[1]));
            *to++ = *iter;
        }
    }
    cls.erase(to,cls.end());
}




bool ConstraintDB::checkSatisfied(Clause& cl) const{
    for(uint32_t i = 0; i != cl.size(); ++i) {
        if(litIsTrue(cl.literals[i])) return true;
    }
    return false;
}


void ConstraintDB::writeDimacs(OFStream& out, bool skip_sat_cls) const{
    for(uint32_t i = 0; i != fix_cls.size(); ++i){
        Clause* cl = fix_cls[i];
        if(skip_sat_cls && checkSatisfied(*cl)) continue;

        cl->print(out);
    }
}

bool ConstraintDB::debugVerifyModel(){


	for(int i(fix_cls.size());i--;){
        if(checkSatisfied(*(fix_cls[i]))) continue;

        if(settings.console_out != NULL){
        	settings.console_out->writeString("|Found unsatisfied clause: ");
            Clause& cl = *(fix_cls[i]);
            for(uint32_t j = 0; j != cl.size(); ++j) {
                Literal lit = cl[j];
                settings.console_out->writeString(" %i: %s\n",lit.toDIMACS(),(litIsFalse(lit)?"false":"unknown"));
            }
        }

        return false;
    }

    for(uint i = 0; i != numberVariables();++i){
    		Variable var(i);
    		assert(!litIsUnknown(var.toPosLit()) || litWatchedClauses(var.toPosLit()).empty() && litWatchedClauses(var.toNegLit()).empty() );

    		Literal true_lit = litIsTrue(var.toPosLit()) ? var.toPosLit() : var.toNegLit();
    		we_vec& watches = litWatchedClauses(true_lit.flipped());

    		for(uint j = 0; j != watches.size(); ++j){
    			WatchedElement& we = watches[j];
    			if(!we.isShort() || litIsTrue(we.getFirstLiteral()) || (we.isTernary() && litIsTrue(we.getSecondLiteral()))) continue;
    			if(settings.console_out != NULL){
    				settings.console_out->writeString("|Found unsatisfied short clause: ");
    				settings.console_out->writeString(" %i: %s\n",true_lit.flipped().toDIMACS(),(litIsFalse(true_lit.flipped())?"false":"unknown"));
    				settings.console_out->writeString(" %i: %s\n",we.getFirstLiteral().toDIMACS(),(litIsFalse(we.getFirstLiteral())?"false":"unknown"));
    			    if(we.isTernary()){
    			    	settings.console_out->writeString(" %i: %s\n",we.getSecondLiteral().toDIMACS(),(litIsFalse(we.getSecondLiteral())?"false":"unknown"));}
    			}
    			return false;
    		}
   }

    if(settings.console_out != NULL)
    	settings.console_out->writeString("|Verified %d original clauses and %d short clauses ",fix_cls.size(), numberShortClses());
    return true;
}




void ConstraintDB::printModel(){
	printf("v ");
	const uint32_t sz = numberVariables();
	for(uint i = 0; i != sz;++i){
		Variable v(i);
		TrBool val = varValue(v);
		assert(val.isKnown());
		if(val.isUnknown()) continue;
		printf("%d ",(val.isFalse()?v.toNegLit().toDIMACS():v.toPosLit().toDIMACS()));
	}
	printf(" 0\n");
}




const char*  constraintdb_cpp_version(){return "ConstraintDB.cpp:   $Revision: 1.12 $";}


