/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#ifndef LAZY_FRACTION_HPP_
#define LAZY_FRACTION_HPP_

#include <stdint.h>
#include <cassert>
#include <cmath>
#include <string>
#include <iostream>
#include <iomanip>


template<class T>
void printHex(T t){
	std::cout << std::hex << t << std::dec << std::endl;
}


/**@param base_log2: denomerator will be 2^base_log2*/
template<uint32_t base_log2, class num_t = uint16_t, class den_t = uint16_t>
class  LazyFraction{

//class part
public:
//memory usage and precision
typedef num_t numerator_t;
typedef den_t denominator_t;
const static uint  num_prec 			= sizeof(num_t)*8;
const static den_t denom_base			= 1 << base_log2;
const static den_t max_denom_log		= ~0;
const static num_t right_bit_set	 	= 1;
const static num_t left_bit_set	  		= right_bit_set << (num_prec-1);
const static num_t max_num_for_extend 	= left_bit_set 	>> base_log2;
const static num_t max_num_for_mult 	= left_bit_set 	>> 1;
const static num_t decrease_mask		= (right_bit_set << base_log2) -1;


private:
#define LAZY_FRAC_INVARIANTS(f)  assert(!(f).extendable() && !(f).leadingBitSet((f).numerator));\
	assert((f).numerator != 0 || ((f).denom_log == LazyFraction<base_log2,num_t,den_t>::max_denom_log))



static bool leadingBitSet(const num_t t){return t & left_bit_set;}


//object part

num_t  numerator;	//numerator of the fraction
den_t  denom_log; //denominator of the fraction = (2^base_log2)^denom_log


bool extendable() 	const{	return numerator != 0 && numerator < max_num_for_extend && denom_log < max_denom_log;}
bool reducable() 	const{ 	return numerator != 0 && !(numerator & decrease_mask) && denom_log > 0;}

void extendFraction(){assert(denom_log < max_denom_log); 	++denom_log;numerator <<= base_log2;}
void reduceFraction(){assert(denom_log != 0);				--denom_log;numerator >>= base_log2;}


public:
	LazyFraction(num_t num, den_t expo_of_denom = 0)
		:numerator(num),denom_log(num == 0? max_denom_log: expo_of_denom){
		assert(!leadingBitSet(numerator));
		while(extendable()){ extendFraction();}
		LAZY_FRAC_INVARIANTS(*this);}

	LazyFraction(const LazyFraction& f)					:numerator(f.numerator),denom_log(f.denom_log){};

	LazyFraction& operator=(const LazyFraction& rh)		{numerator=rh.numerator; denom_log = rh.denom_log;return *this;}

	static LazyFraction 	minNonZeroNumber()			{return LazyFraction(1,max_denom_log);};

	static LazyFraction 	maxNumber()					{return LazyFraction(left_bit_set - 1,0);};

	inline const den_t		getLogOfDenominator()const	{return denom_log;}

	inline const num_t		getNumerator()		const	{return numerator;}


	inline LazyFraction& 	operator+=(const LazyFraction& add);

	inline LazyFraction& 	operator*=(const LazyFraction& factor);

	inline void				weaken(const den_t new_denom_log = max_denom_log );

	inline double 			toDouble() const;

	inline std::string 		toString(bool reduce = true) const;



	template<uint32_t N,class A,class B> friend
	bool operator<(const LazyFraction<N,A,B>& left, const LazyFraction<N,A,B>& right);

	template<uint32_t N,class A,class B> friend
	bool operator<=(const LazyFraction<N,A,B>& left, const LazyFraction<N,A,B>& right);

	template<uint32_t N,class A,class B> friend
	bool operator>(const LazyFraction<N,A,B>& left, const LazyFraction<N,A,B>& right);

	template<uint32_t N,class A,class B> friend
	bool operator>=(const LazyFraction<N,A,B>& left, const LazyFraction<N,A,B>& right);

	template<uint32_t N,class A,class B> friend
	bool operator==(const LazyFraction<N,A,B>& left, const LazyFraction<N,A,B>& right);

};

template<uint32_t base_log2,class num_t, class den_t>
LazyFraction<base_log2,num_t,den_t>& LazyFraction<base_log2,num_t,den_t>::operator+=(const LazyFraction<base_log2,num_t,den_t>& add){
	LAZY_FRAC_INVARIANTS(*this); LAZY_FRAC_INVARIANTS(add);

	if(add.numerator == 0){return *this;}
	while(denom_log > add.denom_log && numerator != 0){assert(denom_log != 0); reduceFraction();}
	if(numerator == 0){ return *this = add;}

	if(denom_log == add.denom_log){
		numerator += add.numerator;
	}else{
		LazyFraction<base_log2,num_t,den_t> add_cp(add);
		while(add_cp.denom_log > denom_log){
			add_cp.reduceFraction();
			if(add_cp.numerator == 0){ return *this;}
		}
		assert(add_cp.denom_log == denom_log);
		assert(denom_log != 0 || (uint64_t)numerator + (uint64_t)add_cp.numerator < (uint64_t)max_num_for_mult);
		numerator += add_cp.numerator;
	}

	if(leadingBitSet(numerator)){reduceFraction();}

	LAZY_FRAC_INVARIANTS(*this);
	return *this;
}

template<uint32_t base_log2,class num_t,class den_t>
LazyFraction<base_log2,num_t,den_t>& LazyFraction<base_log2,num_t,den_t>::operator*=(const LazyFraction<base_log2,num_t,den_t>& factor){
	assert(max_denom_log - denom_log >= factor.denom_log); // can be improved but here it's sufficient (guaranteed by call of weaken)
	LAZY_FRAC_INVARIANTS(*this); LAZY_FRAC_INVARIANTS(factor);

	if(numerator == 0)			{return *this;} //would be handled later but here it's faster
	if(factor.numerator == 0)	{return *this = factor;}

	denom_log 			+= factor.denom_log;
	num_t fact_num 		= factor.numerator;
	uint32_t k 			= num_prec;

	//omit leading zeros
	while(k-- && !leadingBitSet(fact_num)){fact_num <<= 1;}

	//handle leftmost set bit
	num_t add 			= numerator;
	fact_num 			<<= 1;

	//search for more set bits
	while(k-- ){

		if(numerator < max_num_for_mult){//normal iterative mult
			numerator <<= 1;
		}else{
			assert(denom_log > 0); // can be improved but here it's sufficient since guaranteed by call of weaken

			//reduce precision -> divide denominator and numerator by denom_base (=2^base_log2) (possibly with loss)
			--denom_log;
			numerator 	>>= (base_log2-1); //only base_log2-1 since numerator has to be multiplied by 2
			add 		>>= base_log2;
		}

		//if bit is set then add the previous numerator once
		numerator 	+=	((num_t)leadingBitSet(fact_num)) * add;

		//look at next bit
		fact_num <<= 1;
	}

	LAZY_FRAC_INVARIANTS(*this);
	return *this;
}

template<uint32_t base_log2,class num_t,class den_t>
void LazyFraction<base_log2,num_t,den_t>::weaken(const den_t new_denom_log){
	LAZY_FRAC_INVARIANTS(*this);

	if(	new_denom_log <= denom_log || (numerator = ((num_t)(new_denom_log - denom_log))>>1,numerator == 0)){
		numerator = 0;	denom_log = max_denom_log;
		return;
	}

	denom_log 		= new_denom_log;
	while(extendable()){extendFraction();}
	LAZY_FRAC_INVARIANTS(*this);
}



template<uint32_t base_log2,class num_t,class den_t>
double LazyFraction<base_log2,num_t,den_t>::toDouble() const{
	double d_base   = denom_base;
	double denom	= pow(d_base,denom_log);
	return (double)numerator/denom;
}

template<uint32_t base_log2,class num_t,class den_t>
std::string LazyFraction<base_log2,num_t,den_t>::toString(bool reduce) const{
   	char buff[40];
   	LazyFraction cp(*this);
   	if(reduce){	while(cp.reducable()){cp.reduceFraction();}}

   	sprintf(buff,"%u/(%u^%u)",cp.numerator,denom_base,cp.denom_log);
   	return std::string(buff);
}


template<uint32_t base_log2,class num_t,class den_t>
bool operator<(const LazyFraction<base_log2,num_t,den_t>& left, const LazyFraction<base_log2,num_t,den_t>& right){
	LAZY_FRAC_INVARIANTS(left); LAZY_FRAC_INVARIANTS(right);

	return (left.denom_log == right.denom_log) ?
			left.numerator <  right.numerator:
			left.denom_log >  right.denom_log;
}

template<uint32_t base_log2,class num_t,class den_t>
bool operator<=(const LazyFraction<base_log2,num_t,den_t>& left, const LazyFraction<base_log2,num_t,den_t>& right){
	LAZY_FRAC_INVARIANTS(left); LAZY_FRAC_INVARIANTS(right);

	return (left.denom_log == right.denom_log) ?
			left.numerator <= right.numerator:
			left.denom_log >  right.denom_log;
}


template<uint32_t base_log2,class num_t,class den_t>
bool operator>(const LazyFraction<base_log2,num_t,den_t>& left, const LazyFraction<base_log2,num_t,den_t>& right){
	LAZY_FRAC_INVARIANTS(left); LAZY_FRAC_INVARIANTS(right);

	return (left.denom_log == right.denom_log) ?
			left.numerator >  right.numerator:
			left.denom_log <  right.denom_log;
}

template<uint32_t base_log2,class num_t,class den_t>
bool operator>=(const LazyFraction<base_log2,num_t,den_t>& left, const LazyFraction<base_log2,num_t,den_t>& right){
	LAZY_FRAC_INVARIANTS(left); LAZY_FRAC_INVARIANTS(right);


	return (left.denom_log == right.denom_log) ?
			left.numerator >= right.numerator:
			left.denom_log <  right.denom_log;
}


template<uint32_t base_log2,class num_t,class den_t>
bool operator==(const LazyFraction<base_log2,num_t,den_t>& left, const LazyFraction<base_log2,num_t,den_t>& right){
	LAZY_FRAC_INVARIANTS(left); LAZY_FRAC_INVARIANTS(right);

	return left.denom_log == right.denom_log &&  left.numerator == right.numerator;
}


#endif
